#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides the WavRead class for reading wave files.

The WavRead class is returned by wavfile.open() when opening a file in read
mode.
"""

from . import base
from . import chunk
from . import exception


class WavRead(base.Wavfile):
    """Class for reading a wave file"""

    def __init__(self, f):
        """
        Initialise the WavRead object.

        :param f: Either a path to a wave file or a pointer to an open file.
        :type f: Union[str, os.PathLike, typing.IO]
        """
        base.Wavfile.__init__(self)
        self._init_fp(f, 'rb')

        # read the file header
        try:
            self._init_file()
        except BaseException:
            self.close()
            raise

        self.fp.seek(self._data_chunk.content_start)

    def _init_file(self):
        """
        Read the file and initialise the object properties.
        """
        self.fp.seek(0, 2)
        file_size = self.fp.tell()
        self.fp.seek(0)
        fmt_chunk = None

        # Read the file
        while file_size - self.fp.tell() > 0:
            chnk = chunk.Chunk(self.fp)

            # rewind to chunk start
            self.fp.seek(-chunk.Chunk.offset, 1)

            # interpret each chunk
            if chnk.chunk_id == chunk.ChunkID.RIFF_CHUNK:
                self._riff_chunk = chunk.RiffChunk(self.fp)
            elif chnk.chunk_id == chunk.ChunkID.FMT_CHUNK:
                fmt_chunk = chunk.WavFmtChunk(self.fp)
            elif chnk.chunk_id == chunk.ChunkID.DATA_CHUNK:
                if fmt_chunk is None:
                    raise exception.ReadError('DATA chunk read before FMT chunk')
                self._data_chunk = chunk.WavDataChunk(self.fp, fmt_chunk)

            # skip superfluous bytes
            if chnk.chunk_id != chunk.ChunkID.RIFF_CHUNK:
                chnk.skip()

        # go to data chunk content start ready to read samples
        self.fp.seek(self._data_chunk.content_start)

    def __copy__(self):
        """
        Create a shallow copy of the WavRead object.
        """
        newobj = type(self)(self.fp.name)
        newobj.__dict__.update({key: value for key, value in self.__dict__.items()
                                if key not in ('fp', '_riff_chunk', '_data_chunk')})
        newobj.fp.seek(self.fp.tell())
        return newobj

    def _block_iterator(self, method, num_frames):
        """
        Read blocks of frames using an iterator.
        """
        while True:
            audio = getattr(self, method)(num_frames=num_frames)
            if len(audio) > 0:
                yield audio
            else:
                break

    def close(self):
        """
        Close the wav file.
        """
        base.Wavfile.close(self)
        if self._should_close_file:
            self.fp.close()
        self._should_close_file = False

    def read_int(self, num_frames=None):
        """
        Read, at most, num_frames frames from the audio stream in integer format. The method returns
        a list of lists with dimensions (N,C), where C is the number of audio channels. Choosing
        N = None or N < 0 will read all remaining samples.

        :param num_frames: Maximum number of frames to read.
        :type num_frames: int
        :return: The audio samples as a list of lists.
        :rtype: list[list[int]]
        """
        audio = self._data_chunk.read_frames(num_frames)
        if self.format == chunk.WavFormat.IEEE_FLOAT:
            for n in range(len(audio)):
                for m in range(len(audio[n])):
                    audio[n][m] = self._convert_float_to_int(audio[n][m])
        return audio

    def iter_int(self, num_frames=None):
        """
        This method is equivalent to read_int(), except that it returns a generator rather than
        a block of sample.

        :param num_frames: Number of frames to read on each iteration.
        :type num_frames: int
        :return: A generator to yield the next frame(s) of audio.
        :rtype: Iterator[list[list[int]]]
        """
        return self._block_iterator('read_int', num_frames)

    def read_float(self, num_frames=None):
        """
        Read, at most, num_frames frames from the audio stream in float format in the range [-1, 1).
        The method returns a list of lists with size [N][C], where C is the number of audio
        channels. Choosing N = None or N < 0 will read all remaining samples.

        :param num_frames: Maximum number of frames to read.
        :type num_frames: int
        :return: The audio samples as a list of lists.
        :rtype: list[list[float]]
        """
        audio = []
        if self.format == chunk.WavFormat.IEEE_FLOAT:
            audio = self._data_chunk.read_frames(num_frames)
        elif self.format == chunk.WavFormat.PCM:
            audio = self.read_int(num_frames)
            if self._bytes_per_sample == 1:
                convert = self._convert_unsigned_int_to_float
            else:
                convert = self._convert_signed_int_to_float
            for i in range(0, len(audio)):
                for j in range(0, len(audio[0])):
                    audio[i][j] = convert(audio[i][j])

        return audio

    def iter_float(self, num_frames=None):
        """
        This method is equivalent to read_float(), except that it returns a generator rather than
        a block of samples.

        :param num_frames: Number of frames to read on each iteration.
        :type num_frames: int
        :return: A generator to yield the next frame(s) of audio.
        :rtype: Iterator[list[list[float]]]
        """
        return self._block_iterator('read_float', num_frames)

    def read(self, num_frames=None):
        """
        Read, at most, num_frames frames from the audio stream in its native format. The method
        returns a list of lists with dimensions (N,C), where C is the number of audio channels.
        Choosing N = None or N < 0 will read all remaining samples.

        :param num_frames: Maximum number of frames to read.
        :type num_frames: int
        :return: The audio samples as a list of lists.
        :rtype: list[list[Union[int, float]]]
        """
        return self._data_chunk.read_frames(num_frames)

    def iter(self, num_frames=None):
        """
        This method is equivalent to read(), except that it returns a generator rather than
        a block of samples.

        :param num_frames: Number of frames to read on each iteration.
        :type num_frames: int
        :return: A generator to yield the next frame(s) of audio.
        :rtype: Iterator[list[list[Union[int, float]]]]
        """
        return self._block_iterator('read', num_frames)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides the WavWrite class for writing wave files.

The WavWrite class is returned by wavfile.open() when opening a file in write
mode.
"""

from . import base
from . import chunk


class WavWrite(base.Wavfile):
    """Class for writing a wave file"""

    def __init__(self, fp, sample_rate=44100, num_channels=None, bits_per_sample=16,
                 fmt=chunk.WavFormat.PCM):
        """
        Initialise the WavWrite object.

        :param fp: Either a path to a wave file or a pointer to an open file.
        :type fp: Union[str, os.PathLike, typing.IO]
        :param sample_rate: The sample rate for the new file.
        :type sample_rate: int
        :param num_channels: The number of audio channels for the new file.
        :type num_channels: int
        :param bits_per_sample: The number of bits to encode each audio sample.
        :type bits_per_sample: int
        :param fmt: The audio format (chunk.WavFormat.PCM, chunk.WavFormat.IEEE_FLOAT)
        :type fmt: wavfile.chunk.WavFormat
        """
        base.Wavfile.__init__(self)

        self._is_open = True
        self._init_fp(fp, 'wb')

        # initialise each of the riff chunk
        self._riff_chunk = chunk.RiffChunk(self.fp)
        self._riff_chunk.format = chunk.RiffFormat.WAVE
        fmt_chunk = chunk.WavFmtChunk(self.fp)
        fmt_chunk.audio_fmt = fmt
        self._data_chunk = chunk.WavDataChunk(self.fp, fmt_chunk)
        self._data_chunk.fmt_chunk.sample_rate = int(sample_rate)
        if num_channels is not None:
            self._data_chunk.fmt_chunk.num_channels = int(num_channels)
        self._data_chunk.fmt_chunk.bits_per_sample = int(bits_per_sample)

        # go to data chunk content start ready to write samples
        self.fp.seek(self._data_chunk.content_start)

    @staticmethod
    def _data_are_floats(data):
        """
        Check for any floats in data.
        """
        return any([any([isinstance(y, float) for y in x]) for x in data]) or \
            base.Wavfile._buffer_max_abs(data) <= 1.0

    def write(self, audio):
        """
        Write audio data to the file. The data should be a list of lists with dimensions (N,C),
        where N is the number of frames and C is the number of audio channels. The data may be int
        or float. The data may be converted if they do match the format of the destination file.

        :param audio: Audio frames to write.
        :type audio: list[list[Union[int, float]]]
        """
        if self._data_are_floats(audio) and \
                self.format == chunk.WavFormat.PCM:
            # data are floats but we are writing integers
            for n in range(len(audio)):
                for m in range(len(audio[n])):
                    audio[n][m] = self._convert_float_to_int(audio[n][m])
        elif not self._data_are_floats(audio) and \
                self.format == chunk.WavFormat.IEEE_FLOAT:
            # data are integers but we are writing floats
            for n in range(len(audio)):
                for m in range(len(audio[n])):
                    audio[n][m] = self._convert_int_to_float(audio[n][m])

        self._data_chunk.write_frames(audio)

    def close(self):
        """
        Close the file.
        """
        num_align_bytes = self._data_chunk.size % chunk.Chunk.align
        if num_align_bytes > 0:
            self._data_chunk.skip()
            self._data_chunk.write(bytearray(num_align_bytes))
        base.Wavfile.close(self)
        if self._should_close_file:
            self.fp.close()
        self._should_close_file = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Chunk-based helper classes for working with RIFF files.
"""

import struct
import sys
from enum import Enum

from . import exception


class RiffFormat(Enum):
    """RIFF file format"""
    WAVE = b'WAVE'


class ChunkID(Enum):
    """RIFF chunk identifiers"""
    RIFF_CHUNK = b'RIFF'
    FMT_CHUNK = b'fmt '
    DATA_CHUNK = b'data'
    UNKNOWN_CHUNK = b'    '


class WavFormat(Enum):
    """Wav audio data format"""
    PCM = 0x0001
    IEEE_FLOAT = 0x0003


class Chunk:
    """Chunk read and write"""

    align = 2
    offset = 8
    _min_size = 0

    def __init__(self, fp, bigendian=False):
        """
        Initialise the chunk from a file pointer.

        :param fp: Open file pointer.
        :type fp: typing.IO
        :param bigendian: Is the file bid endian?
        :type bigendian: bool
        """
        self.fp = fp
        self.bigendian = bigendian
        if not hasattr(self, 'chunk_id'):
            self.chunk_id = None
        self.size = 0
        self.start = self.fp.tell()
        self._header_is_written = False

        if 'r' in self.fp.mode:
            chunk_id = self.fp.read(4)
            try:
                self.chunk_id = ChunkID(chunk_id)
            except ValueError:
                self.chunk_id = ChunkID.UNKNOWN_CHUNK
            if len(self.chunk_id.value) > 0:
                self.size = self.read_int(4, signed=True)
        else:
            self.write_header()

    @property
    def endianness(self):
        """The endianness in text form."""
        if self.bigendian:
            return 'big'
        else:
            return 'little'

    @property
    def content_start(self):
        """The start position of the chunk content"""
        return self.start + self.offset

    def skip(self):
        """
        Skip to the end of the chunk.
        """
        if self.size == 0:
            raise exception.Error('Chunk has no payload to skip')
        self.fp.seek(self.content_start + self.size)

    def read(self, nbytes):
        """
        Read data from the chunk.

        :param nbytes: The number of bytes to read.
        :type nbytes: int
        """
        data = self.fp.read(nbytes)
        return data

    def write(self, data):
        """
        Write data to the chunk. Adjust the chunk size accordingly.
        """
        data_size = len(data)
        real_size = data_size - (self.size - (self.fp.tell() - self.content_start))
        self.size += max(real_size, 0)
        if not self._header_is_written and 'w' in self.fp.mode and not self.fp.closed:
            pos = self.fp.tell()
            self.write_header()
            self.fp.seek(pos)
        self.fp.write(data)

    def write_header(self):
        """
        Write the chunk header to the file, esp. the updated chunk size.
        """
        self.fp.seek(self.start)
        self.fp.write(self.chunk_id.value)
        self.fp.write(self.int_to_bytes(self.size, 4, signed=False))
        self._header_is_written = True

    def bytes_to_int(self, data, signed=True):
        """
        Read a signed integer from the specified number of bytes of the input
        chunk.

        :param data: Input bytes to interpret as a signed integer.
        :type data: bytes
        :param signed: Is the integer signed?
        :type signed: bool
        :return: The integer value.
        :rtype: int
        """
        # noinspection PyTypeChecker
        return int.from_bytes(data, byteorder=self.endianness, signed=signed)

    def read_int(self, nbytes, signed=True):
        """
        Read an integer from the chunk.

        :param nbytes: Number of bytes encoding the integer.
        :type nbytes: int
        :param signed: Whether the integer is signed.
        :type signed: bool
        :return: The integer value.
        :rtype: int
        """
        data = self.read(nbytes)
        if len(data) < nbytes:
            raise IOError('Could not read enough bytes. Maybe EOF.')
        return self.bytes_to_int(data, signed=signed)

    def int_to_bytes(self, data, nbytes, signed=True):
        """
        Convert an integer to bytes.

        :param data: The integer to encode.
        :type data: int
        :param nbytes: Number of bytes encoding the integer.
        :type nbytes: int
        :param signed: Whether the integer is signed.
        :type signed: bool
        :return: The bytes.
        :rtype: bytes
        """
        # noinspection PyTypeChecker
        return data.to_bytes(nbytes, byteorder=self.endianness, signed=signed)

    def write_int(self, data, nbytes, signed=True):
        """
        Write an integer to the chunk.

        :param data: The integer to encode.
        :type data: int
        :param nbytes: Number of bytes encoding the integer.
        :type nbytes: int
        :param signed: Whether the integer is signed.
        :type signed: bool
        """
        self.write(self.int_to_bytes(data, nbytes, signed=signed))

    def _get_float_struct_fmt(self, nbytes):
        """Get the format specifier for packing and unpacking floats"""

        endianness = '<' if self.endianness == 'little' else '>'
        try:
            return endianness + {
                4: 'f',
                8: 'd',
            }[nbytes]
        except KeyError:
            raise exception.Error('Cannot convert float with the specified bit depth')

    def bytes_to_float(self, data):
        """
        Read a float from the specified number of bytes of the input chunk.

        :param data: Input bytes to interpret as a float.
        :type data: bytes
        :return: The float value.
        :rtype: float
        """
        return struct.unpack(self._get_float_struct_fmt(len(data)), data)[0]

    def read_float(self, nbytes):
        """
        Read a float from the chunk.

        :param nbytes: Number of bytes encoding the float.
        :type nbytes: int
        :return: The float value.
        :rtype: float
        """
        data = self.read(nbytes)

        if len(data) < nbytes:
            raise IOError('Could not read enough bytes. Maybe EOF.')
        return self.bytes_to_float(data)

    def float_to_bytes(self, data, nbytes):
        """
        Convert a float to bytes.

        :param data: The float to encode.
        :type data: float
        :param nbytes: Number of bytes encoding the float.
        :type nbytes: int
        :return: The bytes.
        :rtype: bytes
        """
        return struct.pack(self._get_float_struct_fmt(nbytes), data)

    def write_float(self, data, nbytes):
        """
        Write a float to the chunk.

        :param data: The float to encode.
        :type data: float
        :param nbytes: Number of bytes encoding the float.
        :type nbytes: int
        """
        self.write(self.float_to_bytes(data, nbytes))

    def close(self):
        """
        Close the chunk and update the header.
        """
        if self.size < self._min_size:
            raise exception.WriteError('Required data have not been written to the file')
        if 'w' in self.fp.mode and not self.fp.closed:
            self._header_is_written = False
            self.write_header()


class RiffChunk(Chunk):
    """Riff container chunk read and write"""

    _min_size = 4

    def __init__(self, fp):
        """
        Initialise the chunk from a file pointer.

        :param fp: Open file pointer.
        :type fp: typing.IO
        """
        self.chunk_id = ChunkID.RIFF_CHUNK
        self.format = RiffFormat.WAVE

        Chunk.__init__(self, fp, bigendian=False)

        if 'r' in self.fp.mode:
            if self.chunk_id != ChunkID.RIFF_CHUNK:
                raise exception.ReadError('Chunk is not a RIFF chunk')
            self.format = RiffFormat(self.read(4))
        else:
            self.write(RiffFormat.WAVE.value)

    def close(self):
        """
        Close the chunk and update the header.
        """
        if self.format is None:
            raise exception.WriteError('RIFF format is not set')
        Chunk.close(self)


class WavFmtChunk(Chunk):
    """Wave format chunk read and write"""

    _min_size = 16
    _audio_fmt_size = 2
    _num_channels_size = 2
    _sample_rate_size = 4
    _byte_rate_size = 4
    _block_align_size = 2
    _bits_per_sample_size = 2

    def __init__(self, fp):
        """
        Initialise the chunk from a file pointer.

        :param fp: Open file pointer.
        :type fp: typing.IO
        """
        self.chunk_id = ChunkID.FMT_CHUNK
        self.audio_fmt = WavFormat.PCM
        self.num_channels = 0
        self.sample_rate = 0
        self.byte_rate = 0
        self.block_align = 0
        self.bits_per_sample = 0

        Chunk.__init__(self, fp, bigendian=False)

        if 'r' in self.fp.mode:
            if self.chunk_id != ChunkID.FMT_CHUNK:
                raise exception.ReadError('Chunk is not a FMT chunk')
            self.audio_fmt = WavFormat(self.read_int(self._audio_fmt_size, signed=False))
            self.num_channels = self.read_int(self._num_channels_size, signed=False)
            self.sample_rate = self.read_int(self._sample_rate_size, signed=False)
            self.byte_rate = self.read_int(self._byte_rate_size, signed=False)
            self.block_align = self.read_int(self._block_align_size, signed=False)
            self.bits_per_sample = self.read_int(self._bits_per_sample_size, signed=False)
        else:
            self.write_fmt()

    @property
    def bytes_per_sample(self):
        """Number of bytes per audio sample"""
        return self.bits_per_sample // 8

    @property
    def signed(self):
        """Whether the integer representation is signed"""
        return self.bytes_per_sample > 1

    def write_fmt(self):
        """
        Write the format data to the file.
        """
        self.fp.seek(self.content_start)
        self.write_int(self.audio_fmt.value, self._audio_fmt_size, signed=False)
        self.write_int(self.num_channels, self._num_channels_size, signed=False)
        self.write_int(self.sample_rate, self._sample_rate_size, signed=False)
        self.write_int(self.byte_rate, self._byte_rate_size, signed=False)
        self.write_int(self.block_align, self._block_align_size, signed=False)
        self.write_int(self.bits_per_sample, self._bits_per_sample_size, signed=False)

    def close(self):
        """
        Close the chunk and update the header and format data.
        """
        if 'w' in self.fp.mode and not self.fp.closed:
            self.write_fmt()
        Chunk.close(self)


class WavDataChunk(Chunk):
    """Wave data chunk read and write"""

    def __init__(self, fp, fmt_chunk):
        """
        Initialise the chunk from a file pointer.

        :param fp: Open file pointer.
        :type fp: typing.IO
        :param fmt_chunk: The associated format chunk.
        :type fmt_chunk: WavFmtChunk
        """
        self.chunk_id = ChunkID.DATA_CHUNK
        Chunk.__init__(self, fp, bigendian=False)

        self.__did_warn = False
        self.fmt_chunk = fmt_chunk

        if 'r' in self.fp.mode:
            if self.chunk_id != ChunkID.DATA_CHUNK:
                raise exception.ReadError('Chunk is not a DATA chunk')

    @property
    def num_frames(self):
        """Number of audio frames in the file"""
        try:
            return self.size // self.fmt_chunk.block_align
        except ZeroDivisionError:
            return 0

    def close(self):
        """
        Close the chunk and update the header and format data.
        """
        self.fmt_chunk.close()
        Chunk.close(self)

    def read_sample(self):
        """
        Read a sample from the chunk.
        """
        if self.fmt_chunk.audio_fmt == WavFormat.PCM:
            return self.read_int(self.fmt_chunk.bytes_per_sample, signed=self.fmt_chunk.signed)
        elif self.fmt_chunk.audio_fmt == WavFormat.IEEE_FLOAT:
            return self.read_float(self.fmt_chunk.bytes_per_sample)

    def read_frames(self, nframes=None):
        """
        Read a number of audio frames from the chunk. A frame is a collection of samples, from each
        audio channel, associated with a single time instant.

        :param nframes: Number of audio frames to read.
        :type nframes: int
        :return: The audio frames as a list of lists.
        :rtype: list[list[Union[int, float]]]
        """

        if nframes is None or nframes < 0:
            # read all remaining frames
            nframes = self.num_frames

        # do not try to read past the end
        num_frames = min(nframes, self.num_frames - self.tell())

        audio = []
        for n in range(num_frames):
            frame = []
            for m in range(self.fmt_chunk.num_channels):
                frame.append(self.read_sample())
            audio.append(frame)

        return audio

    @property
    def _max_val(self):
        """Maximum value"""
        if self.fmt_chunk.audio_fmt == WavFormat.PCM:
            if self.fmt_chunk.signed:
                sign_correction = 1
            else:
                sign_correction = 0
            return (2 ** (self.fmt_chunk.bits_per_sample - sign_correction)) - 1
        elif self.fmt_chunk.audio_fmt == WavFormat.IEEE_FLOAT:
            return 1.0

    @property
    def _min_val(self):
        """Minimum value"""
        if self.fmt_chunk.audio_fmt == WavFormat.PCM:
            if self.fmt_chunk.signed:
                return -self._max_val - 1
            else:
                return 0
        elif self.fmt_chunk.audio_fmt == WavFormat.IEEE_FLOAT:
            return -1.0

    def __saturation_warning(self):
        """
        Print a warning to stderr about out-of-range sample values.
        """
        if not self.__did_warn:
            self.__did_warn = True
            print('Saturating one of more sample values that are out-of-range', file=sys.stderr)

    def _saturate(self, val):
        """
        Saturate out-of-range sample values.
        """
        if val > self._max_val:
            val = self._max_val
            self.__saturation_warning()
        elif val < self._min_val:
            val = self._min_val
            self.__saturation_warning()
        return val

    def write_sample(self, sample):
        """
        Write a sample to the chunk.

        :param sample: An audio sample.
        :type sample: Union[int, float]
        """
        sample = self._saturate(sample)
        if self.fmt_chunk.audio_fmt == WavFormat.PCM:
            self.write_int(sample, self.fmt_chunk.bytes_per_sample, signed=self.fmt_chunk.signed)
        elif self.fmt_chunk.audio_fmt == WavFormat.IEEE_FLOAT:
            self.write_float(sample, self.fmt_chunk.bytes_per_sample)

    def write_frames(self, audio):
        """
        Write one or more frames of audio to the chunk.

        :param audio: The audio frames as a list of lists.
        :type audio: list[list[Union[int, float]]]
        """

        num_channels = len(audio[0])
        if num_channels != self.fmt_chunk.num_channels:
            if self.fmt_chunk.num_channels == 0:
                self.fmt_chunk.num_channels = num_channels
            else:
                raise exception.WriteError('Number of channels does not match the format chunk')

        if self.fmt_chunk.block_align == 0:
            self.fmt_chunk.block_align = \
                self.fmt_chunk.num_channels * self.fmt_chunk.bytes_per_sample
            self.fmt_chunk.byte_rate = self.fmt_chunk.block_align * self.fmt_chunk.sample_rate

        for n in range(len(audio)):
            for m in range(len(audio[n])):
                self.write_sample(audio[n][m])

    def seek(self, frame_number, whence=0):
        """
        Move to the specified frame number. The frame positioning mode ``whence`` are: 0 (default) =
        absolute positioning, 1 = relative to current position, 2 = relative to end of last frame.

        :param frame_number: The frame number.
        :type frame_number: int
        :param whence: The frame positioning mode.
        :type whence: int
        :return: The method returns the object.
        :rtype: WavDataChunk
        """
        if frame_number > self.num_frames:
            raise exception.Error('Frame number exceeds number of frames in file')
        relative_pos = self.content_start
        if whence == 0:
            pass
        elif whence == 1:
            relative_pos = self.fp.tell()
        elif whence == 2:
            relative_pos = self.content_start + self.size
        else:
            raise exception.Error('Invalid whence parameter')
        self.fp.seek(relative_pos + (frame_number * self.fmt_chunk.block_align))
        return self

    def tell(self):
        """
        Get the current frame number.
        """
        if self.fmt_chunk.block_align == 0:
            return 0
        else:
            return (self.fp.tell() - self.content_start) // \
                   self.fmt_chunk.block_align



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The main API for reading and writing wav files.
"""

import builtins
import os
from abc import ABC

from . import chunk


class Wavfile(ABC):
    """Abstract base class for wave file read/write"""

    def __init__(self):
        """
        Initialise the base class.
        """
        self.fp = None
        self._should_close_file = False
        self._riff_chunk = None
        self._data_chunk = None

    def _init_fp(self, f, mode):
        """
        Initialise the file pointer.
        """
        # open the file
        if isinstance(f, (str, os.PathLike)):
            self.fp = builtins.open(f, mode)
            self._should_close_file = True
        else:
            self.fp = f

    def _convert_unsigned_int_to_float(self, x):
        """
        Convert unsigned int to float [-1, 1).
        """
        adjust = 2.0 ** (self.bits_per_sample - 1.0)
        return (x - adjust) / adjust

    def _convert_signed_int_to_float(self, x):
        """
        Convert signed int to float [-1, 1).
        """
        return x / (2.0 ** (self.bits_per_sample - 1.0))

    def _convert_int_to_float(self, x):
        """
        Convert int to float [-1, 1).
        """
        if self._data_chunk.fmt_chunk.signed:
            return self._convert_signed_int_to_float(x)
        else:
            return self._convert_unsigned_int_to_float(x)

    def _convert_float_to_unsigned_int(self, x):
        """
        Convert float [-1, 1) to unsigned int.
        """
        return int(round(((x + 1.0) / 2.0) * (2.0 ** self.bits_per_sample)))

    def _convert_float_to_signed_int(self, x):
        """
        Convert float [-1, 1) to signed int.
        """
        return int(round(x * (2.0 ** (self.bits_per_sample - 1.0))))

    def _convert_float_to_int(self, x):
        """
        Convert float [-1, 1) to int.
        """
        if self._data_chunk.fmt_chunk.signed:
            return self._convert_float_to_signed_int(x)
        else:
            return self._convert_float_to_unsigned_int(x)

    @property
    def _bytes_per_sample(self):
        """Number of bytes per audio sample"""
        return self._data_chunk.fmt_chunk.bytes_per_sample

    @property
    def num_channels(self):
        """Number of audio channels in the file"""
        return self._data_chunk.fmt_chunk.num_channels

    @property
    def sample_rate(self):
        """Sampling rate of the audio data"""
        return self._data_chunk.fmt_chunk.sample_rate

    @property
    def bits_per_sample(self):
        """Number of bits per audio sample"""
        return self._data_chunk.fmt_chunk.bits_per_sample

    @property
    def format(self):
        """Audio sample format"""
        return self._data_chunk.fmt_chunk.audio_fmt

    @property
    def num_frames(self):
        """Number of audio frames in the file"""
        try:
            return self._data_chunk.size // self._block_align
        except ZeroDivisionError:
            return 0

    @property
    def _block_align(self):
        """Number of audio frames in the file"""
        return self._data_chunk.fmt_chunk.block_align

    @staticmethod
    def _buffer_max_abs(data):
        """Max(Abs(X)) for an audio buffer"""
        return max([max([abs(y) for y in x]) for x in data])

    def seek(self, frame_number, whence=0):
        """
        Move to the specified frame number. The frame positioning mode ``whence`` are: 0 (default) =
        absolute positioning, 1 = relative to current position, 2 = relative to end of last frame.

        :param frame_number: The frame number.
        :type frame_number: int
        :param whence: The frame positioning mode.
        :type whence: int
        :return: The method returns the object.
        :rtype: Wavfile
        """
        self._data_chunk.seek(frame_number, whence)
        return self

    def tell(self):
        """Get the current frame number."""
        return self._data_chunk.tell()

    def close(self):
        """
        Close the wav file.
        """
        total_size = \
            self._data_chunk.size + chunk.Chunk.offset + \
            self._data_chunk.fmt_chunk.size + chunk.Chunk.offset + \
            4  # riff chunk contains four bytes indicating the format
        self._riff_chunk.size = total_size
        self._data_chunk.close()
        self._riff_chunk.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
wavfile
=======

Read/write wave audio files to/from lists of native Python types.

The library currently supports PCM (integer) and IEEE float formats, and supports arbitrary integer
precision, including: 16-, 24-, 32-, and 64-bit samples.

Usage: reading wave files
-------------------------
::

    f = wavfile.open(file, 'r')

where ``file`` is either a path to a wave file or a pointer to an open file. This returns a
``wavfile.wavread.WavRead`` object with the following properties:

``num_channels``
  The number of audio channels in the stream.

``sample_rate``
  The sampling rate/frequency of the audio stream.

``bits_per_sample``
  The number of bits per audio sample.

``num_frames``
  The total number of audio frames in the audio stream. A frame is a block of samples, one for each
  channel, corresponding to a single sampling point.

``format``
  Audio sample format.

The object also has the following methods:

``read([N])``
  Read, at most, ``N`` frames from the audio stream in their unmodified native format. The method
  returns a list of lists with size ``(N,C)``, where ``C`` is the number of audio channels.
  Excluding ``N``, choosing ``N = None` or ``N < 0`` will read all remaining samples.

``read_int([N])``
  This method is identical to ``read()`` except that it returns the samples as integers using the
  specified bit depth.

``read_float(N)``
  This method is identical to ``read()`` except that it returns the samples as floats in the range
  [-1, 1).

``iter([N])``
  Similar to ``read()`` but used in iterator contexts to read successive groups of audio frames.

``iter_int(N)``
  Similar to ``read_int()`` but used in iterator contexts to read successive groups of audio frames.

``iter_float(N)``
  Similar to ``read_float()`` but used in iterator contexts to read successive groups of audio
  frames.

``seek(N, [W])``
  Move to the ``N``th frame in the audio stream. The position mode can be changed by setting ``W``:
  0 (default) = absolute positioning, 1 = relative to current position, 2 = relative to end of last
  frame.

``tell()``
  Return the current frame in the audio stream.

``close()``
  Close the instance.

Alternatively, the following shortcut function is provided::

    audio, sample_rate, bits_per_sample = wavfile.read(file, fmt='int')

where fmt is ``'int'``, ``'float'``, or ``'native'``; and audio is the audio data. The function
reads all audio data in the file.

Usage: writing wave files
-------------------------
::

    f = wavfile.open(file,
                     'w',
                     sample_rate=44100,
                     num_channels=None,
                     bits_per_sample=16,
                     fmt=chunk.WavFormat.PCM)

where ``sample_rate`` is the sampling rate for the new file, ``num_channels`` is the number of audio
channels, ``bits_per_sample`` is the number of bits used to encode each sample, and ``fmt`` is the
audio sample format. If ``num_channels`` is unspecified it will be determined automatically from the
first block of samples that are written (see below). This returns a ``wavfile.wavread.WavWrite``
object. The object shares its properties with the ``wavfile.wavread.WavRead`` class. The object also
offers the same ``seek()``, ``tell()``, and ``close()`` methods. In addition, the following methods
are provided for writing audio data:

write(N)
  Write ``N`` frames of integers or floats to the audio file. The data should be contained in a list
  of lists with size ``(N,C)``, where ``C`` is the number of audio channels. The data may be ``int``
  or ``float``. The data may be converted if they do match the format of the destination file.

Alternatively, the following shortcut function is provided::

    wavfile.write(file, audio, sample_rate=44100, bits_per_sample=16, fmt=chunk.WavFormat.PCM)

where ``audio`` is the audio data to write to the file.
"""
import os.path

from . import chunk
from . import wavread
from . import wavwrite
from .exception import Error, WriteError
from .version import __VERSION__


def open(f, mode=None, sample_rate=44100, num_channels=None, bits_per_sample=16,
         fmt=chunk.WavFormat.PCM):
    """
    Open the wave file.

    If writing and ``num_channels`` is unspecified, it is determined automatically from the first
    block of samples.

    :param f: Either a path to a wave file or a pointer to an open file.
    :type f: Union[str, os.PathLike, typing.IO]
    :param mode: Open the file for reading ('r', 'rb') or writing ('w', 'wb').
    :type mode: str
    :param sample_rate: The sample rate for the new file (write only).
    :type sample_rate: int
    :param num_channels: The number of audio channels for the new file (write only).
    :type num_channels: int
    :param bits_per_sample: The number of bits to encode each audio sample (write only).
    :type bits_per_sample: int
    :param fmt: The audio format (write only) (chunk.WavFormat.PCM, chunk.WavFormat.IEEE_FLOAT)
    :type fmt: wavfile.chunk.WavFormat
    :return: Returns a wavfile.wavread.WavRead object or wavfile.wavwrite.WavWrite object.
    :rtype: Union[wavfile.wavread.WavRead, wavfile.wavwrite.WavWrite]
    """
    if mode is None:
        if hasattr(f, 'mode'):
            mode = f.mode
        else:
            mode = 'rb'
    if mode in ('r', 'rb'):
        return wavread.WavRead(f)
    elif mode in ('w', 'wb'):
        if not isinstance(fmt, chunk.WavFormat):
            raise WriteError('Invalid format')
        return wavwrite.WavWrite(f, sample_rate=sample_rate, num_channels=num_channels,
                                 bits_per_sample=bits_per_sample, fmt=fmt)
    else:
        raise Error("mode must be 'r', 'rb', 'w', or 'wb'")


def read(path, fmt='int'):
    """
    Shortcut function to read a wave file. The audio can be read directly (``fmt='native'``),
    converted to integers (``fmt='int'``), or converted to floating point (``fmt='float'``).

    :param path: Path to the wave audio file.
    :type path: Union[str, os.PathLike]
    :param fmt: Read the file as 'int', 'float', or 'native'.
    :type fmt: str
    :return: The audio data, the sample rate, and the bit depth.
    :rtype: tuple(list[list[Union[int, float]]], int, int)
    """
    with open(path, 'r') as wf:
        fs = wf.sample_rate
        bits_per_sample = wf.bits_per_sample
        fmt_methods = {
            'int': wf.read_int,
            'float': wf.read_float,
            'native': wf.read,
        }
        audio_data = None
        try:
            audio_data = fmt_methods[fmt]()
        except KeyError:
            raise Error('Unknown format. Options are: ' + ', '.join(fmt_methods.keys()))

    return audio_data, fs, bits_per_sample


def write(path, audio_data, sample_rate=44100, bits_per_sample=16, fmt=chunk.WavFormat.PCM):
    """
    Shortcut function to write a wave file. The data should be contained in a list of lists with
    size (N,C), where C is the number of audio channels. The data may be int or float. The data may
    be converted if they do match the format of the destination file.

    :param path: Path to the newly created wave file.
    :type path: Union[str, os.PathLike]
    :param audio_data: The data to be written to the file.
    :type audio_data: list[list[Union[int, float]]]
    :param sample_rate: The sample rate for the new file.
    :type sample_rate: int
    :param bits_per_sample: The number of bits to encode each audio sample (write only).
    :type bits_per_sample: int
    :param fmt: The audio format (chunk.WavFormat.PCM, chunk.WavFormat.IEEE_FLOAT)
    :type fmt: wavfile.chunk.WavFormat
    """
    with open(path, 'w', sample_rate=sample_rate, bits_per_sample=bits_per_sample, fmt=fmt) as wf:
        wf.write(audio_data)


def split(path):
    """
    Split a multichannel wave file in to multiple mono wave files.

    :param path: Path to the multichannel wave file.
    :type path: Union[str, os.PathLike]
    """

    with open(path, 'r') as wfp:
        num_channels = wfp.num_channels

        # filenames for new files
        filenames = []
        base, ext = os.path.splitext(path)
        for i in range(num_channels):
            filenames.append('{:s}_{:02d}{:s}'.format(base, i, ext))

        # open the wave files for writing
        outs = []
        for i in range(num_channels):
            outs.append(
                wavwrite.WavWrite(
                    filenames[i],
                    sample_rate=wfp.sample_rate,
                    num_channels=1,
                    bits_per_sample=wfp.bits_per_sample,
                    fmt=wfp.format
                )
            )

        # read the input file and write to the output files
        for audio in wfp.iter(1):
            for i in range(num_channels):
                outs[i].write([[audio[0][i]]])

        # close the output files
        for i in range(num_channels):
            outs[i].close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Package exceptions.
"""


class Error(Exception):
    """Generic package error"""
    pass


class ReadError(Error):
    """Package file read error"""
    pass


class WriteError(Error):
    """Package file write error"""
    pass



import pathlib
from setuptools import find_packages, setup

with open('src/wavfile/version.py') as f:
    __VERSION__ = ""
    exec(f.read())

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="wavfile",
    version=__VERSION__,
    description="Read/write wave audio files to/from lists of native Python types.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/chummersone/pywavfile",
    author="Christopher Hummersone",
    author_email="chummersone@users.noreply.github.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[],
    entry_points={},
)
